# -*- coding: utf-8 -*-
"""
Created on Wed Dec 20 19:16:05 2017

@author: Darin
"""

import numpy as np
import scipy as sp
import NURBS as NURBS

def Compliance(Ellist, Node, t, u, du, dvnode, dvdof):
    ''' Computes compliance of the structure and its sensitivity '''
    dim = Node.shape[0]
    x = sp.special.legendre(12).weights[:,0]/2.
    w = sp.special.legendre(12).weights[:,2]
    
    comp = 0
    dcomp = np.zeros(dvnode.size)
    dcpdv = np.zeros((dim,Node.shape[1],dvnode.size))
    for i in range(dvnode.size):
        dcpdv[dvdof[i],dvnode[i],i] = 1.
    for el in Ellist:
        par_length = el.knots[-1] - el.knots[0]
        xi = x*par_length + par_length/2. + el.knots[0]
        eldof = np.vstack((2*el.cp, 2*el.cp+1)).reshape(-1,order='F')
        
        ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
        ny[0,:] *= -1
        ny_raw = ny.copy()
        dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
        ny = ny/dsdxi.reshape((1,-1))
        dxidxi = par_length/2.
        
        p = el.knots.size - el.cp.size - 1
        N = np.kron(NURBS.NURBS_Shape(xi, el.knots, p), np.eye(dim))
        
        comp += np.dot(np.kron(w*dsdxi*dxidxi,np.ones(2)), np.dot(N,t[eldof])*np.dot(N,u[eldof]))
        
        # Sensitivities
        # int(t*u'ds)
        dcomp += np.dot(np.kron(w*dsdxi*dxidxi,np.ones(2)), np.dot(N,t[eldof]).reshape(-1,1)*np.dot(N,du[eldof,:]))
        # int(t*u*ds')
        for i in range(dvnode.size):
            if dvnode[i] in el.cp:
                dny = NURBS.NURBS_Slope(xi, el.knots, dcpdv[:,el.cp,i])[-1::-1,:]
                dny[0,:] *= -1
                ddsdxi = (dny*ny_raw)[-1::-1,:]
                ddsdxi = ddsdxi[dvdof[i],:]/dsdxi

                dcomp[i] += np.dot(np.kron(w*ddsdxi*dxidxi,np.ones(2)), np.dot(N,t[eldof])*np.dot(N,u[eldof]))
            
    return comp, dcomp

def Volume(Ellist, Node, dvnode, dvdof):
    ''' Computes volume of the structure and its sensitivity '''
    dim = Node.shape[0]
    eta = sp.special.legendre(12).weights[:,0]/2.
    w = sp.special.legendre(12).weights[:,2]
    
    vol = 0
    dvol = np.zeros(dvnode.size)
    dcpdv = np.zeros((dim,Node.shape[1],dvnode.size))
    for i in range(dvnode.size):
        dcpdv[dvdof[i],dvnode[i],i] = 1.
    for el in Ellist:
        par_length = el.knots[-1] - el.knots[0]
        xi = eta*par_length + par_length/2. + el.knots[0]
        
        ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
        ny[0,:] *= -1
        ny_raw = ny.copy()
        dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
        ny = ny/dsdxi.reshape((1,-1))
        dxidxi = par_length/2.
        
        p = el.knots.size - el.cp.size - 1
        N = NURBS.NURBS_Shape(xi, el.knots, p)
        
        x = np.dot(Node[:,el.cp],N.T)
        xdotn = np.sum(x*ny,axis=0)
        vol += np.dot(xdotn, w*dsdxi*dxidxi)
        
        # Sensitivities
        for i in range(dvnode.size):
            if dvnode[i] in el.cp:
                dny = NURBS.NURBS_Slope(xi, el.knots, dcpdv[:,el.cp,i])[-1::-1,:]
                dny[0,:] *= -1
                ddsdxi = (dny*ny_raw)[-1::-1,:]
                ddsdxi = ddsdxi[dvdof[i],:]/dsdxi
                dny = (dny*dsdxi - ny_raw*ddsdxi)/dsdxi**2
                dx = np.dot(dcpdv[:,el.cp,i],N.T)
                
                xdotdn = np.sum(x*dny,axis=0)
                dxdotn = np.sum(dx*ny,axis=0)
                dvol[i] += np.dot(xdotdn+dxdotn,w*dsdxi*dxidxi) + np.dot(xdotn, w*ddsdxi*dxidxi)
            
    return vol/dim, dvol/dim

def Perimeter(Ellist, Node, dvnode, dvdof):
    ''' Computes perimeter of the structure and its sensitivity '''
    dim = Node.shape[0]
    eta = sp.special.legendre(12).weights[:,0]/2.
    w = sp.special.legendre(12).weights[:,2]
    
    perim = 0
    dper = np.zeros(dvnode.size)
    dcpdv = np.zeros((dim,Node.shape[1],dvnode.size))
    for i in range(dvnode.size):
        dcpdv[dvdof[i],dvnode[i],i] = 1.
        
    for el in Ellist:
        par_length = el.knots[-1] - el.knots[0]
        xi = eta*par_length + par_length/2. + el.knots[0]
        
        ny = NURBS.NURBS_Slope(xi, el.knots, Node[:,el.cp])[-1::-1,:]
        ny[0,:] *= -1
        ny_raw = ny.copy()
        dsdxi = np.sqrt(np.sum(ny*ny,axis=0))
        ny = ny/dsdxi.reshape((1,-1))
        dxidxi = par_length/2.
        
        perim += np.dot(dsdxi, w*dxidxi)
        
        # Sensitivities
        for i in range(dvnode.size):
            if dvnode[i] in el.cp:
                dny = NURBS.NURBS_Slope(xi, el.knots, dcpdv[:,el.cp,i])[-1::-1,:]
                dny[0,:] *= -1
                ddsdxi = (dny*ny_raw)[-1::-1,:]
                ddsdxi = ddsdxi[dvdof[i],:]/dsdxi
 
                dper[i] += np.dot(ddsdxi, w*dxidxi)
            
    return perim, dper