# -*- coding: utf-8 -*-
"""
Created on Fri Dec 15 11:35:02 2017

@author: Darin
"""

import numpy as np
import Quadrature as quad
import Visualization as viz
import Solver as slv
import Functions as fun
import MMA as mma

# Setting basic properties
mu = 27e9
nu = 0.33
xel = 10
yel = 4
Order = 3
p = Order-1
Elsize = 4
Ellength = 1.
dim = 2

# Deriving quantities
Nel = 2*xel+2*yel
Nnode = Nel*(Elsize-1)
length = Nel*Ellength
spacing = length/(Nnode)

# Initialization
Ellist = np.zeros(Nel,dtype=object)
Node = np.zeros((dim,Nnode))
knots = np.concatenate([np.zeros(p), np.linspace(0.,1.,Elsize-Order+2), np.ones(p)])
Node[:,0] = [-xel*Ellength/2, -yel*Ellength/2]
point = 0
el = 0

#Setting up node and element arrays assuming rectangular initial shape
for i in range(yel):
    cp = [point]
    for i in range(1,Elsize):
        Node[:,point+1] = Node[:,point]
        Node[1,point+1] += spacing
        point += 1
        cp.append(point)
    Ellist[el] = quad.Element(np.array(cp), knots)
    el += 1
for i in range(xel):
    cp = [point]
    for i in range(1,Elsize):
        Node[:,point+1] = Node[:,point]
        Node[0,point+1] += spacing
        point += 1
        cp.append(point)
    Ellist[el] = quad.Element(np.array(cp), knots)
    el += 1
for i in range(yel):
    cp = [point]
    for i in range(1,Elsize):
        Node[:,point+1] = Node[:,point]
        Node[1,point+1] -= spacing
        point += 1
        cp.append(point)
    Ellist[el] = quad.Element(np.array(cp), knots)
    el += 1
for i in range(xel):
    cp = [point]
    for i in range(1,Elsize):
        # Complete the wrapping
        if point == Nnode-1:
            cp.append(0)
        else:
            Node[:,point+1] = Node[:,point]
            Node[0,point+1] -= spacing
            point += 1
            cp.append(point)
    Ellist[el] = quad.Element(np.array(cp), knots)
    el += 1
# Make boundary smooth
el = yel
Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]
el = xel+yel-1
Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[el+1].cp[0]] - Node[:,Ellist[el+1].cp[1]]
el = 2*yel+xel
Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]
el = Nel-1
Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[0].cp[0]] - Node[:,Ellist[0].cp[1]]
# Plot shape
viz.Plot_Shape(Ellist, Node)

# Specify which nodes are subject to optimization
width = Ellength*xel
freeNode = np.where(np.logical_and(Node[0,:] > -width/2+1e-12+Ellength,Node[0,:] < width/2-1e-12-Ellength))[0].astype(int)
freeNode = np.concatenate([[freeNode],[np.ones(freeNode.size,dtype=int)]]).T

# Fixing and freeing dof
u = np.zeros(dim*Nnode)
t = 0*u
fixdof = np.where(Node[0,:] < -width/2+1e-12)[0].astype(int)
fixdof = np.vstack([dim*fixdof, dim*fixdof+1]).reshape(-1,order='F')
freedof = np.setdiff1d(np.arange(dim*Nnode),fixdof)
load = np.where(Node[0,:] > width/2-1e-12)[0]
#t[2*load+1] = 1e6
t[2*load] = 5e4*np.concatenate((np.arange(load.size//2+1), np.arange(load.size//2)[-1::-1]))

# Optimization variables
ndv = freeNode.shape[0]
xold1 = Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]]
xold2 = Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]]
xmin = Ellength*yel/8*np.ones(ndv/2)
xmax =  Ellength*yel*np.ones(ndv/2)
low = xmin.copy()
upp = xmax.copy()
a0 = 1
a = np.ones(1)
c = np.ones(1)
d = np.ones(1)

# Initial plot
viz.Plot_Initial(Ellist, Node, fixdof, 2*load)
    
vol0, dvol   = fun.Volume(Ellist, Node, freeNode[:,0], freeNode[:,1])
vollist = []
complist = []
perimlist = []
comp_min = np.inf
step = 0
for it in range(30):
    # Assemble matrices
    G, H = quad.Assemble_K(Ellist, Node, mu, nu)
    # Create linear system operator and right-hand-side vector
    A, b = slv.Operator(G, H, u, t, freedof, fixdof)
    
    # Iterative solver works poorly due to H and G having drastically different norms
    # Optionally improve conditioning of system for iterative solver
    factor = 1
    H *= factor
    #x, info = sp.sparse.linalg.gmres(A, b)
    x = np.linalg.solve(A, b)
    H /= factor
    u[freedof] = factor*x[freedof]
    t[fixdof] = x[fixdof]
    
#    DShape = Node + width/np.max(np.abs(u))*0.1*u.reshape((-1,2)).T
#    viz.Plot_Shape(Ellist, DShape)
        
    vol, dvol   = fun.Volume(Ellist, Node, freeNode[:,0], freeNode[:,1])
    perim, dper = fun.Perimeter(Ellist, Node, freeNode[:,0], freeNode[:,1])
    vollist.append(vol)
    perimlist.append(perim)
    
    xval = Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]]
    dper = dper[:dper.size//2] - dper[dper.size//2:]
    dvol = dvol[:dvol.size//2] - dvol[dvol.size//2:]
    xmma,ymma,zmma,lam,eta,low,upp = mma.mmasub(it,xval,xmin,xmax,xold1,xold2,perim,dper,vol-0.75*vol0,dvol,low,upp,a0,a,c,d)
    xold2 = xold1
    xold1 = xval
    xval = xmma
    Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]] = xval
    Node[freeNode[ndv//2:,1],freeNode[ndv//2:,0]] = -xval

    # Ensure smoothness
    el = yel
    Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[el+1].cp[0]] - Node[:,Ellist[el+1].cp[1]]
    el = xel+yel-1
    Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]
    el = 2*yel+xel
    Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[el+1].cp[0]] - Node[:,Ellist[el+1].cp[1]]
    el = Nel-1
    Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]

for it in range(60):
    # Assemble matrices
    G, H = quad.Assemble_K(Ellist, Node, mu, nu)
    # Create linear system operator and right-hand-side vector
    A, b = slv.Operator(G, H, u, t, freedof, fixdof)
    
    # Iterative solver works poorly due to H and G having drastically different norms
    # Optionally improve conditioning of system for iterative solver
    factor = 1
    H *= factor
    #x, info = sp.sparse.linalg.gmres(A, b)
    x = np.linalg.solve(A, b)
    H /= factor
    u[freedof] = factor*x[freedof]
    t[fixdof] = x[fixdof]
    
#    DShape = Node + width/np.max(np.abs(u))*0.5*u.reshape((-1,2)).T
#    viz.Plot_Shape(Ellist, DShape)
    
    dG, dH = quad.kernel_Sensitivities(Ellist, Node, mu, nu, freeNode[:,0], freeNode[:,1])
    #dG2, dH2 = quad.FDGH(Ellist, Node, mu, nu, freeNode, 1e-6)
    du = np.zeros((dim*Nnode,dG.shape[2]))
    dt = np.zeros((dim*Nnode,dG.shape[2]))
    for i in range(dG.shape[2]):
        f = np.dot(dG[:,:,i],t) - np.dot(dH[:,:,i],u)
        xadj = np.linalg.solve(A, f)
        du[freedof, i] = factor*xadj[freedof]
        dt[fixdof, i] = xadj[fixdof]
        
    vol, dvol   = fun.Volume(Ellist, Node, freeNode[:,0], freeNode[:,1])
    t[fixdof] = 0
    comp, dcomp = fun.Compliance(Ellist, Node, t, u, du, freeNode[:,0], freeNode[:,1])
    if comp < comp_min:
        best = Node
        step = it
    complist.append(comp)
    vollist.append(vol)
    
    xval = Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]]
    dcomp = dcomp[:dcomp.size//2] - dcomp[dcomp.size//2:]
    dvol = dvol[:dvol.size//2]    - dvol[dvol.size//2:]
    xmma,ymma,zmma,lam,eta,low,upp = mma.mmasub(it,xval,xmin,xmax,xold1,xold2,comp,dcomp,vol-0.75*vol0,dvol,low,upp,a0,a,c,d)
    xold2 = xold1
    xold1 = xval
    xval = xmma
    Node[freeNode[:ndv//2,1],freeNode[:ndv//2,0]] = xval
    Node[freeNode[ndv//2:,1],freeNode[ndv//2:,0]] = -xval

    # Ensure smoothness
    el = yel
    Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[el+1].cp[0]] - Node[:,Ellist[el+1].cp[1]]
    el = xel+yel-1
    Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]
    el = 2*yel+xel
    Node[:,Ellist[el].cp[-2]] = 2*Node[:,Ellist[el+1].cp[0]] - Node[:,Ellist[el+1].cp[1]]
    el = Nel-1
    Node[:,Ellist[el].cp[1]] = 2*Node[:,Ellist[el-1].cp[-1]] - Node[:,Ellist[el-1].cp[-2]]
    
    # Plot shape
    viz.Plot_Shape(Ellist, Node)
    